/***************************************************************************\
  "table.h" - Symbol Table (Phase C)
  Revision 13; 10 Nov 2002 - 21 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht

  References:
    Dragon pp433-437
    MCD 2.1.11; 2.1.11.3
\***************************************************************************/

#ifndef TABLE_H
#define TABLE_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //
// GLOBAL_VALUE_STRING_SIZE

// This value (known as "m") should be a prime number so that
//   if "n" is the average number of symbols in the symbol table,
//   "n / m" is approximately <= 2.  See (Aho pp433-437).
// According to Grune, a good "m" is 1008 (MCD 2.1.11.3)
#define TABLE_HASH_ARRAY_SIZE 17

// This value (known as "alpha") is multiplied by each character
//   in the symbol's key before they are added together.
//   According to (Aho p437), alpha=1 is the least efficient
//   and alpha=65599 is best for 32-bit integers.
// According to Grune, a good "alpha" is 613 (MCD 2.1.11.3)
#define TABLE_HASH_ALPHA 65599

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "global.h"
#include "token.h"
#include "k_conv.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

typedef enum
{
  tNull, tLabel, tBool, tInt, tRat, tStr
} Type;

char *gvsTypeStrings[] =
{
  "<null>", "label", "bool", "int", "rat", "str"
};

typedef enum
{
  repNull, repIdent, repLiteral
} RepresentationType;

char *gvsRepresentationTypeStrings[] =
{
  "<null>", "identifier", "literal"
};

// Symbol //
typedef struct Symbol_tag
{
  RepresentationType rep;
  Type type;
  bool special; // special boolean value for specific uses:
                // TYPE        MEANING
                // ----        -------
                // tLabel      true if already defined

  void *value; // the value held in the symbol:
               // TYPE         VALUE          MEANING
               // ----         -----          -------
               // tLabel       int*           statement index
               // tInt         int*           integer
               // tBool        bool*          boolean
               // tRat         double*        rational
               // tStr         char*          string
               // tNull        (not defined)  (other)

  struct Symbol_tag *next;
  char *key;
} Symbol;

Symbol *sym_new (char *aKey);
Symbol *sym_copy (Symbol *symbol);
void sym_del (Symbol *symbol);
void sym_del_savevalue (Symbol *symbol);
void sym_set (Symbol *symbol, RepresentationType aRep, Type aType, void *aValue);

// Bucket //
typedef struct
{
  Symbol *head;
  Symbol *tail;
} Bucket;

void buc_init (Bucket *bucket);
Symbol *buc_search (Bucket *bucket, char *aKey);
Symbol *buc_enter (Bucket *bucket, char *aKey);

// Table //
typedef struct
{
  Bucket **array;     // hash array
} Table;

Table *tab_new ();
void tab_del (Table *table);
Symbol *tab_search (Table *table, char *aKey);
Symbol *tab_enter (Table *table, char *aKey);

// Hash //
int hash (char *as);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// sym_new
Symbol *sym_new (char *aKey)
{
  Symbol *temp = (Symbol *) KALLOC(sizeof (Symbol));

  temp->rep = repNull;
  temp->type = tNull;
  temp->special = false;
  temp->value = NULL;

  temp->next = NULL;
  temp->key = k_clone(aKey);

  return temp;
}

//// sym_copy
//// Makes a copy of the symbol and returns it. Does not copy
////   the "next" part.
Symbol *sym_copy (Symbol *symbol)
{
  Symbol *temp = (Symbol *) KALLOC(sizeof (Symbol));

  temp->rep = symbol->rep;
  temp->type = symbol->type;

  if ((temp->type == tLabel) || (temp->type == tInt))
  {
    temp->value = k_new_n (*(int *)(symbol->value));
  }
  else if (temp->type == tBool)
  {
    temp->value = k_new_b (*(bool *)(symbol->value));
  }
  else if (temp->type == tRat)
  {
    temp->value = k_new_d (*(double *)(symbol->value));
  }
  else if (temp->type == tStr)
  {
    temp->value = k_clone ((char *)(symbol->value));
  }
  else
  {
    temp->value = NULL;
  }
  
  temp->key = k_clone (symbol->key);

  return temp;
}

//// sym_del
void sym_del (Symbol *symbol)
{
  // symbol may be NULL //
  if (!symbol) return;

  if (symbol->value)
  {
    KFREE(symbol->value);
  }
  k_del (symbol->key);
  KFREE(symbol);
}

//// sym_del_savevalue
//// Deletes the Symbol, but does not delete the value
void sym_del_savevalue (Symbol *symbol)
{
  // symbol may be NULL //
  if (!symbol) return;

  k_del (symbol->key);
  KFREE(symbol);
}

//// sym_set
void sym_set (Symbol *symbol, RepresentationType aRep, Type aType, void *aValue)
{
  symbol->rep = aRep;
  symbol->type = aType;

  // given value is NULL, so set the value as the default for each type //
  if (aValue == NULL)
  {
    // determine the type of the declaration //
    if (aType == tNull) symbol->value = NULL;
    else if (aType == tLabel) symbol->value = k_new_n (0);
    else if (aType == tBool) symbol->value = k_new_b (false);
    else if (aType == tInt) symbol->value = k_new_n (0);
    else if (aType == tRat) symbol->value = k_new_d (0.0);
    else if (aType == tStr) symbol->value = k_clone ("");
  }
  else
  {
    if (aType == tNull) symbol->value = NULL;
    else if (aType == tLabel) symbol->value = k_new_n (*(int *)aValue);
    else if (aType == tBool) symbol->value = k_new_b (*(bool *)aValue);
    else if (aType == tInt) symbol->value = k_new_n (*(int *)aValue);
    else if (aType == tRat) symbol->value = k_new_d (*(double *)aValue);
    else if (aType == tStr) symbol->value = k_clone (aValue);
  }
}

//// buc_init
void buc_init (Bucket *bucket)
{
  bucket->head = NULL;
  bucket->tail = NULL;
}

//// buc_search
//// Returns pointer to the Symbol with the matching
//// symbol key, or NULL if not found
Symbol *buc_search (Bucket *bucket, char *aKey)
{
  Symbol *temp = bucket->head;

  // Loop to find the matching symbol key //
  while (temp)
  {
    if (k_same (temp->key, aKey))
    {
      return temp;
    }
    else
    {
      temp = temp->next;
    }
  }

  // Not found; return NULL //
  return NULL;
}

//// buc_enter
//// Returns pointer to the Symbol; If not found, it will be created
Symbol *buc_enter (Bucket *bucket, char *aKey)
{
  Symbol *temp;

  temp = buc_search (bucket, aKey);

  // Not found; create and enter it //
  if (temp == NULL)
  {
    temp = sym_new (aKey);

    // First element //
    if (bucket->head == NULL)
    {
      bucket->head = bucket->tail = temp;
    }
    // Other elements //
    else
    {
      bucket->tail->next = temp;
      bucket->tail = temp;
    }
  }

  // Found //
  else
  {
  }

  // Return //
  return temp;
}

//// tab_new
Table *tab_new ()
{
  int i;
  Table *temp = (Table *) KALLOC(sizeof (Table));

  // Initialize the hash array to NULL //
  temp->array = (Bucket **) KALLOC(sizeof (Bucket *) * TABLE_HASH_ARRAY_SIZE);
  for (i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    temp->array[i] = NULL;
  }

  return temp;
}

//// tab_del
void tab_del (Table *table)
{
  int i;

  // free each element in array //
  for (i = 0; i < TABLE_HASH_ARRAY_SIZE; i++)
  {
    KFREE(table->array[i]);
  }
  KFREE(table->array);

  KFREE(table);
}

//// tab_search
//// Returns pointer to the node, or NULL if not found
Symbol *tab_search (Table *table, char *aKey)
{
  Bucket *pBucket;
  Symbol *pSymbol;

  pBucket = table->array[hash (aKey)];

  // There is no Bucket at that hash; Not found //
  if (pBucket == NULL)
  {
    return NULL;
  }
  // Bucket found at that hash //
  else
  {
    pSymbol = buc_search (pBucket, aKey);
    return pSymbol;
  }
}

//// tab_enter
//// Returns pointer to the node; If not found, it will be created
Symbol *tab_enter (Table *table, char *aKey)
{
  int nHash;
  Bucket *pBucket;
  Symbol *pSymbol;

  nHash = hash (aKey);
  pBucket = table->array[nHash];

  // There is no Bucket at that hash; Create Bucket //
  if (pBucket == NULL)
  {
    pBucket = (Bucket *) KALLOC(sizeof (Bucket));
    buc_init (pBucket);
    table->array[nHash] = pBucket;
  }

  // Create Symbol //
  pSymbol = buc_enter (pBucket, aKey);

  return pSymbol;
}

//// hash
//// hash function for the "Alpha" algorithm (Aho p435)
int hash (char *as)
{
  int i;
  int len;
  int total;

  len = strlen (as);
  total = 0;

  for (i = 0; i < len; i++)
  {
    total += as[i] * TABLE_HASH_ALPHA;
  }

  return (total % TABLE_HASH_ARRAY_SIZE);
}

#endif
